@@ -1234,7 +1234,7 @@ subroutine bstarx(params, constants, workspace, x, y)
                     & one, y(:,isph), 1)
             end do
         end do
-    else 
+    else
         !$omp parallel do default(none) shared(params,constants,workspace,x,y) &
         !$omp private(isph) schedule(static,1)
         do isph = 1, params % nsph
@@ -1300,39 +1300,6 @@ subroutine bx(params, constants, workspace, x, y)
     if (constants % dodiag) y = y + x
 end subroutine bx
 
-!> Subroutine used for the GMRES solver with complete matrix
-!! @param[in]      x : Input vector
-!! @param[in, out] y : y=T*x
-!!
-subroutine tx(params, constants, workspace, x, y)
-    implicit none
-    type(ddx_params_type), intent(in) :: params
-    type(ddx_constants_type), intent(in) :: constants
-    type(ddx_workspace_type), intent(inout) :: workspace
-    real(dp), dimension(constants % nbasis, params % nsph, 2), intent(in) :: x
-    real(dp), dimension(constants % nbasis, params % nsph, 2), intent(out) :: y
-
-    real(dp), dimension(constants % nbasis, params % nsph, 2) :: temp_vector
-    y = zero
-    temp_vector = zero
-
-    ! Compute AXr
-    ! call LXr
-    call lx(params, constants, workspace, x(:,:,1), temp_vector(:,:,1))
-    ! Remove the scaling factor
-    call convert_ddcosmo(params, constants, 1, temp_vector(:,:,1))
-    y(:,:,1) = y(:,:,1) + temp_vector(:,:,1)
-    ! Compute BXe
-    call bx(params, constants, workspace, x(:,:,2), temp_vector(:,:,2))
-    y(:,:,2) = y(:,:,2) + temp_vector(:,:,2)
-
-    ! Reset temp_vector to zero
-    temp_vector = zero
-    ! Call CX
-    call cx(params, constants, workspace, x, temp_vector)
-    y = y + temp_vector
-end subroutine tx
-
 !> Subroutine used for the GMRES solver with complete matrix adjoint system
 !! @param[in]      x : Input vector
 !! @param[in, out] y : y=Tstar*x
@@ -1383,37 +1350,27 @@ subroutine prec_tstarx(params, constants, workspace, x, y)
     type(ddx_workspace_type), intent(inout) :: workspace
     real(dp), intent(in) :: x(constants % nbasis, params % nsph, 2)
     real(dp), intent(inout) :: y(constants % nbasis, params % nsph, 2)
-    integer :: n_iter, info
+    integer :: n_iter
     real(dp) :: r_norm
     real(dp), dimension(params % maxiter) :: x_rel_diff
 
     y(:,:,1) = x(:,:,1)
     call convert_ddcosmo(params, constants, 1, y(:,:,1))
     n_iter = params % maxiter
-    if (params % itersolver .eq. 1) then 
-        call jacobi_diis(params, constants, workspace, constants % inner_tol, y(:,:,1), &
-            & workspace % ddcosmo_guess, n_iter, x_rel_diff, lstarx, ldm1x, hnorm, info)
-    else
-        call gmresr(params, constants, workspace, constants % inner_tol, y(:,:,1), &
-            & workspace % ddcosmo_guess, n_iter, r_norm, lstarx, info)
-    end if
-    if (info.ne.0) then
-        write(*,*) 'prec_tstarx: [1] ddCOSMO failed to converge'
-        stop 1
+    call jacobi_diis(params, constants, workspace, constants % inner_tol, y(:,:,1), &
+        & workspace % ddcosmo_guess, n_iter, x_rel_diff, lstarx, ldm1x, hnorm)
+    if (workspace % error_flag.ne.0) then
+        workspace % error_message = 'prec_tstarx: ddCOSMO failed to converge'
+        return
     end if
     y(:,:,1) = workspace % ddcosmo_guess
 
     n_iter = params % maxiter
-    if (params % itersolver .eq. 1) then
-        call jacobi_diis(params, constants, workspace, constants % inner_tol, x(:,:,2), workspace % hsp_guess, &
-            & n_iter, x_rel_diff, bstarx, bx_prec, hnorm, info)
-    else
-        call gmresr(params, constants, workspace, constants % inner_tol, x(:,:,2), workspace % hsp_guess, &
-           & n_iter, r_norm, bstarx, info)
-    end if
-    if (info.ne.0) then
-        write(*,*) 'prec_tstarx: [1] HSP failed to converge'
-        stop 1
+    call jacobi_diis(params, constants, workspace, constants % inner_tol, x(:,:,2), workspace % hsp_guess, &
+        & n_iter, x_rel_diff, bstarx, bx_prec, hnorm)
+    if (workspace % error_flag.ne.0) then
+        workspace % error_message = 'prec_tstarx: HSP failed to converge'
+        return
     end if
     y(:,:,2) = workspace % hsp_guess
 
@@ -1432,22 +1389,17 @@ subroutine prec_tx(params, constants, workspace, x, y)
     type(ddx_workspace_type), intent(inout) :: workspace
     real(dp), intent(in) :: x(constants % nbasis, params % nsph, 2)
     real(dp), intent(inout) :: y(constants % nbasis, params % nsph, 2)
-    integer :: n_iter, info
+    integer :: n_iter
     real(dp) :: r_norm
     real(dp), dimension(params % maxiter) :: x_rel_diff
 
     ! perform A^-1 * Yr
     n_iter = params % maxiter
-    if (params % itersolver .eq. 1) then
-        call jacobi_diis(params, constants, workspace, constants % inner_tol, x(:,:,1), &
-            & workspace % ddcosmo_guess, n_iter, x_rel_diff, lx, ldm1x, hnorm, info)
-    else
-        call gmresr(params, constants, workspace, constants % inner_tol, x(:,:,1), &
-            & workspace % ddcosmo_guess, n_iter, r_norm, lx, info)
-    end if
-    if (info.ne.0) then
-        write(*,*) 'prec_tx: [1] ddCOSMO failed to converge'
-        stop 1
+    call jacobi_diis(params, constants, workspace, constants % inner_tol, x(:,:,1), &
+        & workspace % ddcosmo_guess, n_iter, x_rel_diff, lx, ldm1x, hnorm)
+    if (workspace % error_flag.ne.0) then
+        workspace % error_message = 'prec_tx: ddCOSMO failed to converge'
+        return
     end if
 
     ! Scale by the factor of (2l+1)/4Pi
@@ -1456,18 +1408,13 @@ subroutine prec_tx(params, constants, workspace, x, y)
 
     ! perform B^-1 * Ye
     n_iter = params % maxiter
-    if (params % itersolver .eq. 1) then
-        call jacobi_diis(params, constants, workspace, constants % inner_tol, x(:,:,2), workspace % hsp_guess, &
-            & n_iter, x_rel_diff, bx, bx_prec, hnorm, info)
-    else
-        call gmresr(params, constants, workspace, constants % inner_tol, x(:,:,2), workspace % hsp_guess, &
-         & n_iter, r_norm, bx, info)
-    end if
+    call jacobi_diis(params, constants, workspace, constants % inner_tol, x(:,:,2), workspace % hsp_guess, &
+        & n_iter, x_rel_diff, bx, bx_prec, hnorm)
     y(:,:,2) = workspace % hsp_guess
 
-    if (info.ne.0) then
-        write(*,*) 'prec_tx: [1] HSP failed to converge'
-        stop 1
+    if (workspace % error_flag.ne.0) then
+        workspace % error_message = 'prec_tx: HSP failed to converge'
+        return
     end if
 end subroutine prec_tx
 !
@@ -1529,7 +1476,8 @@ subroutine cstarx(params, constants, workspace, x, y)
         workspace % tmp_sph = zero
         ! Do FMM operations adjointly
         call tree_m2p_bessel_adj(params, constants, constants % lmax0, &
-            & one, workspace % tmp_grid, zero, params % lmax, workspace % tmp_sph)
+            & one, workspace % tmp_grid, zero, params % lmax, &
+            & workspace % tmp_sph)
         call tree_l2p_bessel_adj(params, constants, one, &
             & workspace % tmp_grid, zero, workspace % tmp_node_l)
         call tree_l2l_bessel_rotation_adj(params, constants, &
