@@ -21,13 +21,6 @@ use omp_lib
 ! Disable implicit types
 implicit none
 
-! Interface for the printing function
-interface
-    subroutine print_func_interface(string)
-        character(len=255), intent(in) :: string
-    end subroutine
-end interface
-
 !> Type to check and store user input parameters
 type ddx_params_type
     !> Model to use 1 for COSMO, 2 for PCM, 3 for LPB.
@@ -47,22 +40,11 @@ type ddx_params_type
     integer :: lmax
     !> Number of Lebedev grid points on each sphere.
     integer :: ngrid
-    !> Iterative solver to be used. 1 for Jacobi/DIIS, 2 for GMRES.
-    !!
-    !! Other solvers might be added later.
-    integer :: itersolver
     !> Maximum number of iterations for the iterative solver.
     integer :: maxiter
     !> Number of extrapolation points for Jacobi/DIIS solver. Referenced only
     !!      if Jacobi solver is used.
     integer :: jacobi_ndiis
-    !> Number of last vectors that GMRESR works with. Referenced only if GMRESR
-    !!      solver is used.
-    integer :: gmresr_j
-    !> Dimension of the envoked GMRESR. In case of 0 GMRESR becomes the GCR
-    !!      solver, one of the simplest versions of GMRESR. Referenced only if
-    !!      GMRESR solver is used.
-    integer :: gmresr_dim
     !> Enable (1) or disable (0) use of FMM techniques.
     integer :: fmm
     !> Maximal degree of spherical harmonics for a multipole expansion.
@@ -91,10 +73,17 @@ type ddx_params_type
     integer :: error_flag = 2
     !> Last error message
     character(len=255) :: error_message
-    !> Error printing function
-    procedure(print_func_interface), pointer, nopass :: print_func
     !> integer matvecmem. Build hsp matrix to speed up matrix-vec product
     integer :: matvecmem
+    !> variable to enable debug printins:
+    !> fname: name of the output file
+    character(len=255) :: output_filename
+    !> len_fname: actual length of the output file
+    integer :: len_output_filename
+    !> verbose: true if printing is enabled
+    logical :: verbose
+    !> output unit
+    integer :: iunit
 end type ddx_params_type
 
 contains
@@ -114,8 +103,6 @@ contains
 !! @param[in] ngrid: Number of Lebedev grid points `ngrid` >= 0.
 !! @param[in] matvecmem: handling of sparse matrices. 1 for precomputing them 
 !!      and keeping them in memory, 0 for direct matrix-vector products.
-!! @param[in] itersolver: Iterative solver to be used. 1 for Jacobi iterative
-!!      solver. Other solvers might be added later.
 !! @param[in] maxiter: Maximum number of iterations for an iterative solver.
 !!      maxiter > 0.
 !! @param[in] jacobi_ndiis: Number of extrapolation points for Jacobi/DIIS solver.
@@ -137,17 +124,16 @@ contains
 !! @param[in] charge: Charges of atoms. Dimension is `(nsph)`.
 !! @param[in] csph: Coordinates of atoms. Dimension is `(3, nsph)`.
 !! @param[in] rsph: Van-der-Waals radii of atoms. Dimension is `(nsph)`.
-!! @param[in] print_func: Function to print errors.
+!! @param[in] output_filename: file name of log file.
 !! @param[out] params: Object containing all inputs.
-!! @param[out] info: flag of succesfull exit
 !!      = 0: Succesfull exit
 !!      = -1: One of the arguments had an illegal value, check
 !!          params % error_message
 !!      = 1: Allocation of memory to copy geometry data failed.
 subroutine params_init(model, force, eps, kappa, eta, se, lmax, ngrid, &
-        & matvecmem, itersolver, maxiter, jacobi_ndiis, gmresr_j, gmresr_dim, &
+        & matvecmem, maxiter, jacobi_ndiis, &
         & fmm, pm, pl, nproc, nsph, charge, &
-        & csph, rsph, print_func, params, info)
+        & csph, rsph, output_filename, params)
     !! Inputs
     ! Model to use 1 for COSMO, 2 for PCM, 3 for LPB.
     integer, intent(in) :: model
@@ -169,21 +155,11 @@ subroutine params_init(model, force, eps, kappa, eta, se, lmax, ngrid, &
     ! handling of sparse matrix. 1 for precomputing them and keeping them in
     ! memory, 0 for assembling the mvps on-the-fly.
     integer, intent(in) :: matvecmem
-    ! Iterative solver to be used. 1 for Jacobi/DIIS, 2 for GMRES.
-    ! Other solvers might be added later.
-    integer, intent(in) :: itersolver
     ! Maximum number of iterations for the iterative solver.
     integer, intent(in) :: maxiter
     ! Number of extrapolation points for Jacobi/DIIS solver. Referenced only
     !      if Jacobi solver is used.
     integer, intent(in) :: jacobi_ndiis
-    ! Number of last vectors that GMRESR works with. Referenced only if GMRESR
-    !      solver is used.
-    integer, intent(in) :: gmresr_j
-    ! Dimension of the envoked GMRESR. In case of 0 GMRESR becomes the GCR
-    !      solver, one of the simplest versions of GMRESR. Referenced only if
-    !      GMRESR solver is used.
-    integer, intent(in) :: gmresr_dim
     ! Enable (1) or disable (0) use of FMM techniques.
     integer, intent(in) :: fmm
     ! Maximal degree of spherical harmonics for a multipole expansion.
@@ -204,21 +180,31 @@ subroutine params_init(model, force, eps, kappa, eta, se, lmax, ngrid, &
     real(dp), intent(in) :: csph(3, nsph)
     ! Array of radii of atoms of a dimension (nsph).
     real(dp), intent(in) :: rsph(nsph)
-    ! Error printing function
-    procedure(print_func_interface) :: print_func
+    ! log file name
+    character(len=255) :: output_filename
     !! Outputs
     type(ddx_params_type), intent(out) :: params
-    integer, intent(out) :: info
     !! Local variables
-    integer :: igrid, i
+    integer :: igrid, i, info
     character(len=255) :: string
     !! The code
+    ! Clear error state
+    params % error_flag = 0
+    params % error_message = ''
+    ! parse the log file name
+    if (len(trim(output_filename)) .ne. 0) then
+        params % output_filename = output_filename
+        params % verbose = .true.
+        params % len_output_filename = len(trim(output_filename))
+    else
+        params % output_filename = ''
+        params % verbose = .false.
+        params % len_output_filename = 0
+    end if
     ! Model, 1=COSMO, 2=PCM, 3=LPB
     if ((model .lt. 1) .or. (model .gt. 3)) then
         params % error_flag = 1
         params % error_message = "params_init: invalid value of `model`"
-        call print_func(params % error_message)
-        info = -1
         return
     end if
     params % model = model
@@ -226,8 +212,6 @@ subroutine params_init(model, force, eps, kappa, eta, se, lmax, ngrid, &
     if ((force .lt. 0) .or. (force .gt. 1)) then
         params % error_flag = 1
         params % error_message = "params_init: invalid value of `force`"
-        call print_func(params % error_message)
-        info = -1
         return
     end if
     params % force = force
@@ -235,8 +219,6 @@ subroutine params_init(model, force, eps, kappa, eta, se, lmax, ngrid, &
     if (eps .le. one) then
         params % error_flag = 1
         params % error_message = "params_init: invalid value of `eps`"
-        call print_func(params % error_message)
-        info = -1
         return
     end if
     params % eps = eps
@@ -244,8 +226,6 @@ subroutine params_init(model, force, eps, kappa, eta, se, lmax, ngrid, &
     if ((model .eq. 3) .and. (kappa .le. zero)) then
         params % error_flag = 1
         params % error_message = "params_init: invalid value of `kappa`"
-        call print_func(params % error_message)
-        info = -1
         return
     end if
     params % kappa = kappa
@@ -253,8 +233,6 @@ subroutine params_init(model, force, eps, kappa, eta, se, lmax, ngrid, &
     if ((eta .lt. zero) .or. (eta .gt. one)) then
         params % error_flag = 1
         params % error_message = "params_init: invalid value of `eta`"
-        call print_func(params % error_message)
-        info = -1
         return
     end if
     params % eta = eta
@@ -262,8 +240,6 @@ subroutine params_init(model, force, eps, kappa, eta, se, lmax, ngrid, &
     if ((se .lt. -one) .or. (se .gt. one)) then
         params % error_flag = 1
         params % error_message = "params_init: invalid value of `se`"
-        call print_func(params % error_message)
-        info = -1
         return
     end if
     params % se = se
@@ -271,8 +247,6 @@ subroutine params_init(model, force, eps, kappa, eta, se, lmax, ngrid, &
     if (lmax .lt. 0) then
         params % error_flag = 1
         params % error_message = "params_init: invalid value of `lmax`"
-        call print_func(params % error_message)
-        info = -1
         return
     end if
     params % lmax = lmax
@@ -287,26 +261,13 @@ subroutine params_init(model, force, eps, kappa, eta, se, lmax, ngrid, &
     if (igrid .eq. 0) then
         params % error_flag = 1
         params % error_message = "params_init: Unsupported value of `ngrid`"
-        call print_func(params % error_message)
-        info = -1
         return
     end if
     params % ngrid = ngrid
-    ! Iterative solver: 1=Jacobi 2=GMRES
-    if ((itersolver .lt. 1) .or. (itersolver .gt. 2)) then
-        params % error_flag = 1
-        params % error_message = "params_init: invalid value of `itersolver`"
-        call print_func(params % error_message)
-        info = -1
-        return
-    end if
-    params % itersolver = itersolver
     ! Maximum number of iterations
     if (maxiter .le. 0) then
         params % error_flag = 1
         params % error_message = "params_init: invalid value of `maxiter`"
-        call print_func(params % error_message)
-        info = -1
         return
     end if
     params % maxiter = maxiter
@@ -314,38 +275,13 @@ subroutine params_init(model, force, eps, kappa, eta, se, lmax, ngrid, &
     if (jacobi_ndiis .lt. 0) then
         params % error_flag = 1
         params % error_message = "params_init: invalid value of `jacobi_ndiis`"
-        call print_func(params % error_message)
-        info = -1
         return
     end if
     params % jacobi_ndiis = jacobi_ndiis
-    ! Number of last vectors that GMRESR works with. Referenced only if GMRESR
-    !      solver is used.
-    if (gmresr_j .lt. 1) then
-        params % error_flag = 1
-        params % error_message = "params_init: invalid value of `gmresr_j`"
-        call print_func(params % error_message)
-        info = -1
-        return
-    end if
-    params % gmresr_j = gmresr_j
-    ! Dimension of the envoked GMRESR. In case of 0 GMRESR becomes the GCR
-    !      solver, one of the simplest versions of GMRESR. Referenced only if
-    !      GMRESR solver is used.
-    if (gmresr_dim .lt. 0) then
-        params % error_flag = 1
-        params % error_message = "params_init: invalid value of `gmresr_j`"
-        call print_func(params % error_message)
-        info = -1
-        return
-    end if
-    params % gmresr_dim = gmresr_dim
     ! Check if FMM-acceleration is needed
     if ((fmm .lt. 0) .or. (fmm .gt. 1)) then
         params % error_flag = 1
         params % error_message = "params_init: invalid value of `fmm`"
-        call print_func(params % error_message)
-        info = -1
         return
     end if
     params % fmm = fmm
@@ -357,8 +293,6 @@ subroutine params_init(model, force, eps, kappa, eta, se, lmax, ngrid, &
         if (pm .lt. -1) then
             params % error_flag = 1
             params % error_message = "params_init: invalid value of `pm`"
-            call print_func(params % error_message)
-            info = -1
             return
         end if
         ! Maximal degree of local spherical harmonics. Value -1 means no 
@@ -367,8 +301,6 @@ subroutine params_init(model, force, eps, kappa, eta, se, lmax, ngrid, &
         if (pl .lt. -1) then
             params % error_flag = 1
             params % error_message = "params_init: invalid value of `pl`"
-            call print_func(params % error_message)
-            info = -1
             return
         end if
         ! If far-field interactions are to be ignored
@@ -390,8 +322,6 @@ subroutine params_init(model, force, eps, kappa, eta, se, lmax, ngrid, &
     if (nproc .lt. 0) then
         params % error_flag = 1
         params % error_message = "params_init: invalid value of `nproc`"
-        call print_func(params % error_message)
-        info = -1
         return
     else if (nproc .eq. 0) then
         params % nproc = 1
@@ -403,8 +333,6 @@ subroutine params_init(model, force, eps, kappa, eta, se, lmax, ngrid, &
     if (nsph .le. 0) then
         params % error_flag = 1
         params % error_message = "params_init: invalid value of `nsph`"
-        call print_func(params % error_message)
-        info = -1
         return
     end if
     params % nsph = nsph
@@ -415,8 +343,6 @@ subroutine params_init(model, force, eps, kappa, eta, se, lmax, ngrid, &
         params % error_flag = 1
         params % error_message = "params_init: `charge`, `csph` and `rsph` " &
             & // "allocations failed"
-        call print_func(params % error_message)
-        info = 1
         return
     end if
     params % charge = charge
@@ -427,94 +353,23 @@ subroutine params_init(model, force, eps, kappa, eta, se, lmax, ngrid, &
     else
         params % error_flag = 1
         params % error_message = "params_init: invalid value of `matvecmem`"
-        call print_func(params % error_message)
-        info = -1
         return
     end if
-    ! Set print function for errors
-    params % print_func => print_func
-    ! Clear error state
-    info = 0
-    params % error_flag = 0
-    params % error_message = ""
+    ! init log
+    call init_printing(params)
 end subroutine params_init
 
-!> Print header with parameters
-subroutine params_print(params)
-    !! Input
-    type(ddx_params_type), intent(in) :: params
-    !! Local variables
-    character(len=255) :: string
-    !! The code
-    if (params % error_flag .eq. 1) then
-        string = "params_print: `params` is in error state"
-        call params % print_func(string)
-    else if (params % error_flag .eq. 1) then
-        string = "params_print: `params` is not initialised"
-        call params % print_func(string)
-    else
-        write(string, "(A)") "DDX parameters object"
-        call params % print_func(string)
-        write(string, "(A,A)") "Model: ", trim(model_str(params % model))
-        call params % print_func(string)
-        write(string, "(A,L1)") "Forces: ", params % force .eq. 1
-        call params % print_func(string)
-        write(string, "(A,ES23.16E3)") "Dielectric permittivity (eps): ", &
-            & params % eps
-        call params % print_func(string)
-        write(string, "(A,ES23.16E3)") "Debye-Huckel parameter (kappa): ", &
-            & params % kappa
-        call params % print_func(string)
-        write(string, "(A,ES23.16E3)") "Regularization (eta): ", params % eta
-        call params % print_func(string)
-        write(string, "(A,ES23.16E3)") "Shift of regularization (se): ", &
-            & params % se
-        call params % print_func(string)
-        write(string, "(A,I0)") "Modeling harmonics degree (lmax): ", &
-            & params % lmax
-        call params % print_func(string)
-        write(string, "(A,I0)") "Number of Lebedev grid points (ngrid): ", &
-            & params % ngrid
-        call params % print_func(string)
-        write(string, "(A,A)") "Iterative solver: ", &
-            & trim(itersolver_str(params % itersolver))
-        call params % print_func(string)
-        write(string, "(A,I0)") "maxiter: ", params % maxiter
-        call params % print_func(string)
-        write(string, "(A,I0)") "jacobi_ndiis: ", params % jacobi_ndiis
-        call params % print_func(string)
-        write(string, "(A,I0)") "fmm: ", params % fmm
-        call params % print_func(string)
-        write(string, "(A,I0)") "pm: ", params % pm
-        call params % print_func(string)
-        write(string, "(A,I0)") "pl: ", params % pl
-        call params % print_func(string)
-        write(string, "(A,I0)") "nproc: ", params % nproc
-        call params % print_func(string)
-        write(string, "(A,I0)") "nsph: ", params % nsph
-        call params % print_func(string)
-    end if
-end subroutine params_print
-
 !> Free memory used by parameters
 !! @param[inout] params: Object containing all inputs
 !! @param[out] info: flag of succesfull exit
 !!      = 0: Succesfull exit
 !!      = -1: params is in error state
 !!      = 1: Deallocation of memory failed.
-subroutine params_deinit(params, info)
+subroutine params_deinit(params)
     !! Input
     type(ddx_params_type), intent(inout) :: params
-    !! Output
-    integer, intent(out) :: info
+    integer :: info
     !! Code
-    ! Check if input params is in proper state
-    if (params % error_flag .eq. 1) then
-        params % error_flag = 1
-        params % error_message = "params_deinit: `params` is in error state"
-        info = -1
-        return
-    end if
     ! Deallocate memory to avoid leaks
     deallocate(params % charge, stat=info)
     if (info .ne. 0) then
@@ -560,57 +415,72 @@ subroutine closest_supported_lebedev_grid(ngrid)
     ngrid = ng0(igrid)
 end subroutine
 
-!> Print error message and exit with provided error code
-subroutine error(code, message)
-    integer, intent(in) :: code
-    character(len=*), intent(in) :: message
-    write(0, "(A,A)") "ERROR: ", message
-    write(0, "(A,I2)") "CODE:  ", code
-    stop -1
-end subroutine
-
-
-!> Default printing function
-subroutine print_func_default(string)
-    character(len=255), intent(in) :: string
-!   print "(A)", string
-    write(6,"(A)") trim(string)
-end subroutine
-
-subroutine params_free(params, info)
+subroutine params_free(params)
     implicit none
     type(ddx_params_type), intent(out) :: params
-    integer, intent(out) :: info
     integer :: istat
 
     istat = 0
-    info = 0
+
+    call finalize_printing(params)
+    if (params % error_flag .ne. 0) return
 
     if (allocated(params % charge)) then
         deallocate(params % charge, stat=istat)
         if (istat .ne. 0) then
-            info = 1
-            write(6, *) "`charge` deallocation failed!"
-            stop 1
+            params % error_message = "`charge` deallocation failed!"
+            params % error_flag = 1
+            return
         end if
     end if
     if (allocated(params % csph)) then
         deallocate(params % csph, stat=istat)
         if (istat .ne. 0) then
-            info = 1
-            write(6, *) "`csph` deallocation failed!"
-            stop 1
+            params % error_message = "`csph` deallocation failed!"
+            params % error_flag = 1
+            return
         end if
     end if
     if (allocated(params % rsph)) then
         deallocate(params % rsph, stat=istat)
         if (istat .ne. 0) then
-            info = 1
-            write(6, *) "`rsph` deallocation failed!"
-            stop 1
+            params % error_message = "`rsph` deallocation failed!"
+            params % error_flag = 1
+            return
         end if
     end if
 end subroutine params_free
 
-end module ddx_parameters
+!> Open the log file.
+subroutine init_printing(params)
+    implicit none
+    type(ddx_params_type), intent(inout) :: params
+    logical :: exists
+    if (.not.params % verbose) return
+    inquire(file=params % output_filename(1:params % len_output_filename), &
+        & exist=exists)
+    if (exists) then
+        params % error_message = 'Log file already present'
+        params % error_flag = 1
+        return
+    else
+        params % iunit = 100
+        open(params % iunit, &
+            & file=params % output_filename(1: params % len_output_filename), &
+            & form='formatted')
+    end if
+end subroutine init_printing
 
+!> Close the log file.
+subroutine finalize_printing(params)
+    implicit none
+    type(ddx_params_type), intent(out) :: params
+    if (.not.params % verbose) return
+    close(params % iunit)
+    params % verbose = .false.
+    params % output_filename = ''
+    params % len_output_filename = 0
+    params % iunit = 0
+end subroutine finalize_printing
+
+end module ddx_parameters
